/** \file adu.c The main functions used by all modes of operation. */
#include "adu.h"
#include <dirent.h> /* readdir() */
#include <pwd.h>
#include "format.h"
#include "user.h"
#include "select.cmdline.h"
#include "select.h"
#include "cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"

DEFINE_ERRLIST;
int osl_errno;

/** In case a signal is received, its number is stored here. */
static int signum;

/** Command line and config file options. */
struct gengetopt_args_info conf;

/** Options passed to --select-options. */
struct select_args_info select_conf;


/**
 * The table containing the directory names and statistics.
 */
struct osl_table *dir_table = NULL;

static struct osl_column_description dir_table_cols[] = {
	[DT_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = 0,
		.name = "dir",
	},
	[DT_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "num",
		.compare_function = uint64_compare,
		.data_size = sizeof(uint64_t)
	},
	[DT_PARENT_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "parent_num",
		.compare_function = size_compare,
		.data_size = sizeof(uint64_t)
	},
	[DT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[DT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	}
};

static struct osl_table_description dir_table_desc = {
	.name = "dir_table",
	.num_columns = NUM_DT_COLUMNS,
	.flags = 0,
	.column_descriptions = dir_table_cols,
};

/**
 * The log function.
 *
 * \param ll Loglevel.
 * \param fmt Usual format string.
 *
 * All XXX_LOG() macros use this function.
 */
__printf_2_3 void __log(int ll, const char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[255] = "";

	if (ll < conf.loglevel_arg)
		return;
	outfd = stderr;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, sizeof(str), "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static void close_dir_table(void)
{
	int ret;

	if (!dir_table)
		return;
	NOTICE_LOG("closing dir table\n");
	ret = osl(osl_close_table(dir_table, OSL_MARK_CLEAN));
	if (ret < 0)
		ERROR_LOG("failed to close dir table: %s\n", adu_strerror(-ret));
	free((char *)dir_table_desc.dir);
	dir_table = NULL;
}

void close_all_tables(void)
{
	close_dir_table();
	close_user_tables();
}

static void signal_handler(int s)
{
	signum = s;
}

void check_signals(void)
{
	if (likely(!signum))
		return;
	EMERG_LOG("caught signal %d\n", signum);
	close_all_tables();
	exit(EXIT_FAILURE);
}

static int init_signals(void)
{
	if (signal(SIGINT, &signal_handler) == SIG_ERR)
		return -E_SIGNAL_SIG_ERR;
	if (signal(SIGTERM, &signal_handler) == SIG_ERR)
		return -E_SIGNAL_SIG_ERR;
	if (signal(SIGPIPE, &signal_handler) == SIG_ERR)
		return -E_SIGNAL_SIG_ERR;
	return 1;
}

int open_dir_table(int create)
{

	if (dir_table)
		return 1;
	dir_table_desc.dir = adu_strdup(conf.database_dir_arg);

	if (create) {
		NOTICE_LOG("creating dir table\n");
		int ret = osl(osl_create_table(&dir_table_desc));
		if (ret < 0) {
			free((char *)dir_table_desc.dir);
			return ret;
		}
	}
	INFO_LOG("opening dir table\n");
	return osl(osl_open_table(&dir_table_desc, &dir_table));
}

static int check_args(void)
{
	if (conf.create_given && !conf.base_dir_given)
		return -E_SYNTAX;

	/* remove trailing slashes from base-dir arg */
	if (conf.base_dir_given) {
		size_t len = strlen(conf.base_dir_arg);
		for (;;) {
			if (!len) /* empty string */
				return -ERRNO_TO_ERROR(EINVAL);
			if (!--len) /* length 1 is always OK */
				break;
			if (conf.base_dir_arg[len] != '/')
				break; /* no trailing slash, also OK */
			conf.base_dir_arg[len] = '\0';
		}
	}
	return 1;
}

static int print_complete_help_and_die(void)
{
	const char **line;

	printf("%s-%s\n", CMDLINE_PARSER_PACKAGE, CMDLINE_PARSER_VERSION);
	printf("%s\n\n", gengetopt_args_info_purpose);
	printf("%s\n\n", gengetopt_args_info_usage);

	if (conf.help_given)
		line = gengetopt_args_info_help;
	else
		line = gengetopt_args_info_detailed_help;
	for (; *line; line++)
		printf("%s\n", *line);

	if (conf.help_given)
		line = select_args_info_help;
	else
		line  = select_args_info_detailed_help;
	printf("Select options:\n");
	for (; *line; line++)
		printf("%s\n", *line);

	printf("Interactive commands:\n");
	print_interactive_help();
	cmdline_parser_free(&conf);
	select_cmdline_parser_free(&select_conf);
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
	int ret;
	struct cmdline_parser_params params = {
		.override = 1,
		.initialize = 1,
		.check_required = 0,
		.check_ambiguity = 0,
		.print_errors = 0
	};
	select_cmdline_parser_init(&select_conf);
	cmdline_parser_init(&conf);
	/* ignore errors and print complete help if --help was given */
	cmdline_parser_ext(argc, argv, &conf, &params);
	if (conf.help_given || conf.detailed_help_given)
		print_complete_help_and_die();
	cmdline_parser_free(&conf);
	params.check_required = 1;
	params.check_ambiguity = 1;
	params.print_errors = 1;
	ret = cmdline_parser_ext(argc, argv, &conf, &params);
	if (ret)
		exit(EXIT_FAILURE);

	ret = check_args();
	if (ret < 0)
		goto out;
	ret = init_signals();
	if (ret < 0)
		goto out;
	ret = -E_SYNTAX;
	if (conf.select_given)
		ret = com_select();
	else if (conf.create_given)
		ret = com_create();
	else
		ret = com_interactive();
	if (ret < 0)
		goto out;
out:
	close_all_tables();
	if (ret < 0) {
		ERROR_LOG("%s\n", adu_strerror(-ret));
		return -EXIT_FAILURE;
	}
	cmdline_parser_free(&conf);
	select_cmdline_parser_free(&select_conf);
	return EXIT_SUCCESS;
}
